#property version   "1.0"
#property strict
#property copyright "(c)2021 Edgar Akhmadeev"
#property link      "https://www.mql5.com/en/users/dali"
// 2021.03.23



#include <WinAPI\sysinfoapi.mqh>



struct UUID {
	ulong	l;
	ulong	h;
};

union UUID16 {
	uchar	b[16];
	UUID	d;
};



#define SMBIOS_fields	30



//const uint	SMBIOS_signature = 'RSMB';
const uint 		SMBIOS_signature = ('R' << 24) + ('S' << 16) + ('M' << 8) + 'B';

const string	SMBIOS_labels[19] = {
	"System Management BIOS Version",
	
	"System BIOS Version",
	"Embedded Controller Firmware Version",
	
	"BIOS Vendor",
	"BIOS Version",
	"BIOS Release Date",
	
	"UUID",
	
	"System Manufacturer",
	"Product Name",
	"Product Version",
	"Product Serial Number",
	"SKU Number",
	"Family",
	
	"Mainboard Manufacturer",
	"Mainboard Name",
	"Mainboard Version",
	"Mainboard Serial Number",
	"Asset Tag",
	"Location in chassis"
};



class SMBIOS {
private:
	static string	info[SMBIOS_fields];
	
public:
	static bool		Read();
	static void		Print();
	
	static string	GetUUID()		{ return info[6]; }
	static string	GetProductSN()	{ return info[10]; }
	static string	GetBoardSN()	{ return info[16]; }
};



static string	SMBIOS::info[SMBIOS_fields];



//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
static bool
SMBIOS::Read() {
	// Query size of SMBIOS info.
	uint size = GetSystemFirmwareTable(SMBIOS_signature, 0, NULL, 0);
	
	// Allocate memory for SMBIOS info
	uchar data[];
	if (ArrayResize(data, size) <= 0)
		return false;
	
	// Retrieve the SMBIOS table
	uint written = GetSystemFirmwareTable(SMBIOS_signature, 0, data, size);
	if (written != size)
		return false;
	
	int cnt = 0;
	uint idx = 0;
	info[cnt++] = StringFormat("%d.%d", data[idx + 1], data[idx + 2]);
	idx += 8;
	
	// Process the SMBIOS info
	
	UUID16 uuid;
	string s;
	
	while (idx < size) {
		uchar type = data[idx];
		uchar len = data[idx + 1];
		if (len < 4)
			break;
		
		if (type == 0) {
			info[cnt++] = StringFormat("%d.%d", data[idx + 20], data[idx + 21]);
			info[cnt++] = StringFormat("%d.%d", data[idx + 22], data[idx + 23]);
			idx += len;
		}
		
		else if (type == 1 && len >= 25) {
			idx += 8;	// offset to UUID
			
			// check if there is a valid UUID (not all 0x00 or all 0xff)
			bool all_zero = true, all_one = true;
			for (int i = 0; i < 16 && (all_zero || all_one); i++) {
				if (data[idx + i] != 0x00)
					all_zero = false;
				if (data[idx + i] != 0xFF)
					all_one = false;
			}
		
			if (!all_zero && !all_one) {
				uint idx2 = 0;
				s = "{";
				// As off version 2.6 of the SMBIOS specification, the first 3 fields of the UUID are supposed to be encoded on little-endian. (para 7.2.1)
				s += StringFormat("%x", uuid.b[idx2++] = data[idx + 3]);
				s += StringFormat("%x", uuid.b[idx2++] = data[idx + 2]);
				s += StringFormat("%x", uuid.b[idx2++] = data[idx + 1]);
				s += StringFormat("%x", uuid.b[idx2++] = data[idx + 0]);
				s += "-";
				s += StringFormat("%x", uuid.b[idx2++] = data[idx + 5]);
				s += StringFormat("%x", uuid.b[idx2++] = data[idx + 4]);
				s += "-";
				s += StringFormat("%x", uuid.b[idx2++] = data[idx + 7]);
				s += StringFormat("%x", uuid.b[idx2++] = data[idx + 6]);
				s += "-";
				for (int i = 8; i < 10; i++)
					s += StringFormat("%x", uuid.b[idx2++] = data[idx + i]);
				s += "-";
				for (int i = 10; i < 16; i++)
					s += StringFormat("%x", uuid.b[idx2++] = data[idx + i]);
				s += "}";
				info[cnt++] = s;
			}
			
			idx += 16;		// Skip uuid bytes
			idx += 3;		// Skip unknown info
		}
		
		else
			idx += len;		// Skip formatted area
		
		if (cnt >= 18)
			break;			// Finish after main info
		
		do {
			int i = 0;
			while (idx + i < size && data[idx + i] != 0)
				++i;
			info[cnt++] = CharArrayToString(data, idx, i);
			idx += i + 1;
		} while (data[idx] != 0);
		++idx;
	}
	
	return true;
}



static void
SMBIOS::Print() {
	int size = ArraySize(SMBIOS_labels);
	
	//for (int i = 0; i < size; ++i)
	//	::Print(SMBIOS_labels[i], ": ", info[i]);
	
	for (int i = 0; i < SMBIOS_fields; ++i) {
		if (info[i] == NULL)
			return;
		if (i < size)
			::Print(SMBIOS_labels[i], ": ", info[i]);
		else
			::Print(info[i]);
	}
}
